import json
import os
import shutil
import zipfile
from typing import Optional

import geopandas as gpd
import shapefile
from shapely.geometry import shape, Point, Polygon, mapping
# https://pypi.org/project/pyshp/#the-writer-class

from ..GeoServer import GeoServer
from .ZIP import unpack_zip


def get_shp_data(zip_path, geometry=None, rec_name=None) -> Optional[list]:

    originalRoot, _ = os.path.split(zip_path)
    unpack_zip(zip_path=zip_path, new_path=originalRoot)

    temp_shp_path = os.path.join(originalRoot, os.path.basename(zip_path))
    sf = shapefile.Reader(temp_shp_path)
    records = sf.records()
    shapes = sf.shapes()

    results = []
    for index, (shape_rec, record) in enumerate(zip(shapes, records)):
        shapely_geom = shape(shape_rec.__geo_interface__)

        if isinstance(geometry, Point):
            if geometry.within(shapely_geom):  # within是检查geometry是否完全位于shapely_geom内部
                results.append({'geometry': json.dumps(mapping(geometry)),
                                'data': record[rec_name]})
        elif isinstance(geometry, Polygon):
            if geometry.intersects(shapely_geom):  # 用intersects方法来检查两个几何对象是否相交
                results.append({'geometry': json.dumps(mapping(shapely_geom.intersection(geometry))),
                                'data': record[rec_name]})

    if results:
        return results
    else:
        return None


def get_feature_types(shpPath) -> list:
    """
    获取 SHP 文件中的要素类型。

    :param shpPath: SHP 文件的路径
    :return: 要素类型列表
    """
    # 读取 SHP 文件
    gdf = gpd.read_file(shpPath)
    # 获取要素类型
    feature_types = gdf.geometry.geom_type.unique().tolist()

    return feature_types


def shp_to_zip(shp_path, new_zip_path):
    """
    将shp文件转换为zip文件，并保存到指定路径。
    """

    zip_file = zipfile.ZipFile(new_zip_path, 'w', zipfile.ZIP_STORED)  # 新建zip
    for root, dirs, files in os.walk(shp_path):
        for f in files:
            file_path = os.path.join(root, f)
            zip_file.write(file_path, os.path.basename(file_path))  # 写入
    zip_file.close()  # 保存且关闭
    return True


def create_shp_prj(shp_file_path):
    gdf = gpd.GeoDataFrame(columns=['geometry'], crs='epsg:4326')
    gdf['geometry'] = gpd.GeoSeries(None, name='geometry', crs=gdf.crs)
    gdf.to_file(shp_file_path)
    os.remove(shp_file_path.replace('.shp', '.dbf'))
    os.remove(shp_file_path.replace('.shp', '.shx'))
    os.remove(shp_file_path.replace('.shp', '.cpg'))
    os.remove(shp_file_path)


def write_shp_fields(shapeType, file_path, fields):
    """
    写入shape文件
    :param shapeType: 要素类型
                      NULL = 0；       POINT = 1；       POLYLINE = 3；       POLYGON = 5；        MULTIPOINT = 8；
                      POINTZ = 11；    POLYLINEZ = 13；  POLYGONZ = 15；      MULTIPOINTZ = 18；   POINTM = 21；
                      POLYLINEM = 23； POLYGONM = 25；   MULTIPOINTM = 28；   MULTIPATCH = 31
    :param file_path: shp文件路径
    :param fields: 字段配置列表，列表中每个元素代表一个字段。每个元素是一个列表，依次是字段名，字段类型，字段长度
                   Field name: the name describing the data at this column index.
                   Field type: the type of data at this column index. Types can be:
                               "C": Characters, text.
                               "N": Numbers, with or without decimals.
                               "F": Floats (same as "N").
                               "L": Logical, for boolean True/False values.
                               "D": Dates.
                               "M": Memo, has no meaning within a GIS and is part of the xbase spec instead.
                   Field length: the length of the data found at this column index.
                                 Older GIS software may truncate this length to 8 or 11 characters for "Character" fields.
                   Decimal length: the number of decimal places found in "Number" fields.
    :return: 工作空间列表
    """
    writer = shapefile.Writer(file_path, shapeType=shapeType)
    writer.autoBalance = 1
    for field_name, field_type, decimal in fields:
        writer.field(field_name, field_type, decimal=decimal)
    return writer


def write_shp_geometry(writer: shapefile.Writer, coordinates, attributes, geo_type):
    if geo_type == 'Point':
        writer.point(*coordinates)
    elif geo_type == 'Polygon':
        writer.poly(coordinates)

    writer.record(*attributes.values())
    return writer


def publish_shp(zip_path=None, workspace=None, layer=None, geoserver=None) -> bool:
    """
    发布shp文件
    :param zip_path： shp文件的压缩包路径
    :param workspace: 工作空间名称
    :param layer: 图层名称
    :param geoserver: ip_port、username和password配置的字典
    :return: 是否发布成功
    """

    """工作区管理"""
    # 查询现有workspace列表
    workspace_names = GeoServer.get_workspace_list(geoserver)
    if not workspace_names:
        print(f'workspace列表获取失败')
        return False

    # 如果工作区不存在，则创建workspace
    if workspace not in workspace_names:
        if not GeoServer.create_workspace(geoserver, workspace):
            print(f'{workspace}工作区不存在，创建workspace失败')
            return False
    else:
        print(f'{workspace}工作区已存在')

    """开始处理geoserver发布"""
    zip_name = os.path.basename(zip_path).replace('.zip', '')
    if not layer:
        layer = zip_name

    originalRoot, _ = os.path.split(zip_path)

    output_path = os.path.join(originalRoot, "temp", zip_name)
    unpack_zip(zip_path, output_path)

    # 获取文件夹中的所有文件
    files = [f for f in os.listdir(output_path) if os.path.isfile(os.path.join(output_path, f))]
    # 遍历文件
    for filename in files:
        # 分离文件名和扩展名
        name, ext = os.path.splitext(filename)
        # 构建新的文件名
        new_filename = f"{layer}{ext}"
        # 完整的旧文件路径
        old_filepath = os.path.join(output_path, filename)
        # 完整的新文件路径
        new_filepath = os.path.join(output_path, new_filename)
        # 重命名文件
        try:
            os.rename(old_filepath, new_filepath)
        except Exception as e:
            print(f"重命名文件失败：{e}")
            return False

    temp_ZIP_path = os.path.join(originalRoot, "temp", zip_name + '.zip')
    shp_to_zip(shp_path=output_path, new_zip_path=temp_ZIP_path)
    shutil.rmtree(output_path)

    if GeoServer.publish_shp_with_zip(geoserver, workspace, layer, temp_ZIP_path):
        return True


def find_shp_files(directory):
    """
    在给定的目录中查找所有shp文件
    """
    shp_files = []
    # os.walk()返回一个生成器，它会遍历directory及其子目录
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file.endswith('.shp'):
                # os.path.join()用于构建完整的文件路径
                shp_files.append(os.path.join(root, file))
    return shp_files


def shp_to_geojsonList(shp_file_path):
    # 读取Shapefile
    try:
        sf = shapefile.Reader(shp_file_path)
    except UnicodeDecodeError:
        sf = shapefile.Reader(shp_file_path, encoding='gbk')

    geojsonList = []

    for i in range(len(sf)):
        # 获取每个形状的几何数据
        geometry_shape = sf.shape(i)
        try:
            geojsonList.append(mapping(geometry_shape))
        except Exception as e:
            print(f"shp转geojson失败：{e}")
    return geojsonList



import copy
from typing import List, Optional

import requests


def get_workspace_list(geoserver: dict) -> Optional[List[str]]:
    """
    获取工作空间列表
    :param geoserver： ip_port、username和password配置的字典
    :return: 工作空间列表
    """
    response = requests.get(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces',
        auth=(geoserver["username"], geoserver["password"]),
    )
    # 检查请求是否成功
    if response.status_code == 200:
        # 如果你知道响应是JSON格式，可以这样处理
        workspaces = response.json()['workspaces']['workspace']
        workspace_names = [wp['name'] for wp in workspaces]
        return workspace_names
    else:
        return None


def create_workspace(geoserver: dict, workspace_name: str) -> bool:
    """
    创建工作空间
    :param geoserver： ip_port、username和password配置的字典
    :param workspace_name: 工作空间名称
    :return: 是否创建成功
    """
    response = requests.post(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces',
        headers={'Content-Type': 'application/json'},
        json={
            "workspace": {
                "name": workspace_name
            }
        },
        auth=(geoserver["username"], geoserver["password"]),
    )
    if response.status_code == 201:
        print(f'工作区{workspace_name}创建成功')
        return True
    else:
        print(f'工作区{workspace_name}创建失败')
        return False


def publish_tif(geoserver: dict, workspace: str, layer: str, tif_path: str, fileData: bytes = None) -> bool:
    """
    发布tif文件
    :param geoserver： ip_port、username和password配置的字典
    :param workspace: 工作空间名称
    :param layer: 图层名称
    :param tif_path: tif文件路径
    :param fileData: tif文件数据
    :return: 是否发布成功
    """
    if fileData:
        file_data = fileData
    else:
        file = open(tif_path, 'rb')
        file_data = file.read()
        file.close()

    delete_datastore(geoserver, workspace, layer)
    delete_coveragestore(geoserver, workspace, layer)

    response = requests.put(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/coveragestores/{layer}/file.geotiff',
        headers={'Content-Type': 'image/tiff'},
        data=file_data,
        auth=(geoserver["username"], geoserver["password"]),
    )
    if response.status_code == 201 or response.status_code == 200:
        print('geoserver发布成功')
        return True
    else:
        return False


def publish_shp_with_zip(geoserver: dict, workspace: str, layer: str, zip_path: str) -> tuple[bool, str]:
    """
    发布shp文件
    :param geoserver: ip_port、username和password配置的字典
    :param workspace: 工作空间名称
    :param layer: 图层名称
    :param zip_path: shp文件路径
    :return: shp文件是否成功发布的布尔值。如果发布失败，则返回具体的错误信息；若成功，则为空字符串
    """
    file = open(zip_path, 'rb')
    file_data = file.read()

    delete_datastore(geoserver, workspace, layer)
    delete_coveragestore(geoserver, workspace, layer)

    response = requests.put(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/datastores/{layer}/file.shp',
        headers={'Content-type': 'application/zip'},
        data=file_data,
        auth=(geoserver["username"], geoserver["password"]),
    )
    if response.status_code == 201 or response.status_code == 200:  # 发布请求成功
        state, data = judge_publish_state(geoserver, workspace, layer)  # 判断图层启用状态
        if state is False:  # 图层未启用
            if data.get('srs') is None:
                if reset_CRS_and_publish(geoserver, workspace, layer, data):
                    return True, ""
            else:
                return False, "数据入库成功但发布服务失败，可能由于错误的坐标系定义"
        else:  # 图层启用
            return True, ""
    else:  # 发布请求成功
        return False, "数据入库成功但发布服务失败，未知的原始数据错误"


def judge_publish_state(geoserver: dict, workspace: str, layer: str) -> tuple[bool, dict]:
    """
    判断shp文件发布状态
    :param geoserver: ip_port、username和password配置的字典
    :param workspace: 工作空间名称
    :param layer: 图层名称
    :return: shp文件是否成功发布的布尔值；存储仓库属性值字典
    """
    response = requests.get(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/datastores/{layer}/featuretypes/{layer}',
        auth=(geoserver["username"], geoserver["password"]),
        headers={"Accept": "application/json"}
    )
    data = response.json()
    featureType = data.get('featureType', None)
    if featureType and featureType.get('enabled'):
        return True, featureType
    else:
        return False, featureType


def reset_CRS_and_publish(geoserver: dict, workspace: str, layer: str, data: dict) -> bool:
    # https://docs.geoserver.org/stable/en/api/#1.0.0/featuretypes.yaml

    body = copy.deepcopy(data)
    body.update({"srs": "EPSG:4326",
                 "enabled": True})
    response = requests.put(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/datastores/{layer}/featuretypes/{layer}',
        auth=(geoserver["username"], geoserver["password"]),
        headers={'Content-Type': 'application/json'},
        json={'featureType': body},
        params={'recalculate': ['nativebbox', 'latlonbbox']}
    )
    if response.status_code == 201 or response.status_code == 200:
        return True
    else:
        return False


def delete_datastore(geoserver: dict, workspace: str, datastore: str) -> bool:
    """
    删除数据存储
    :param geoserver： ip_port、username和password配置的字典
    :param workspace: 工作空间名称
    :param datastore: 数据存储名称
    :return: 是否删除成功
    """
    name_list = get_datastores(geoserver, workspace)
    if name_list and datastore in name_list:
        response = requests.delete(
            f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/datastores/{datastore}?recurse=true',
            auth=(geoserver["username"], geoserver["password"]))
        if response.status_code == 200:
            print(f'{datastore}删除成功')
            return True
        else:
            return False
    else:
        print(f'{datastore}：datastore不存在')
        return True


def get_datastores(geoserver: dict, workspace: str) -> Optional[List[str]]:
    """
    获取数据存储列表
    :param geoserver： ip_port、username和password配置的字典
    :param workspace: 工作空间名称
    :return: 数据存储列表
    """
    response = requests.get(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/datastores',
        auth=(geoserver["username"], geoserver["password"]),
    )
    if response.status_code == 200:
        response_json = response.json()['dataStores']
        if response_json:
            datastores = response_json['dataStore']
            datastores_names = [str(i['name']) for i in datastores]
            return datastores_names
        else:
            return None
    else:
        print(f'{workspace}：datastores查询失败')
        return None


def delete_coveragestore(geoserver: dict, workspace: str, coveragestore: str) -> bool:
    """
    删除数据存储
    :param geoserver： ip_port、username和password配置的字典
    :param workspace: 工作空间名称
    :param coveragestore: 数据存储名称
    :return: 是否删除成功
    """
    name_list = get_coveragestores(geoserver, workspace)

    if name_list and coveragestore in name_list:
        response = requests.delete(
            f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/coveragestores/{coveragestore}?purge=all&recurse=true',
            auth=(geoserver["username"], geoserver["password"]))
        if response.status_code == 200:
            print(f'{coveragestore}：coveragestore删除成功')
            return True
        else:
            return False
    else:
        print(f'{coveragestore}：coveragestore不存在')
        return True


def get_coveragestores(geoserver: dict, workspace: str) -> Optional[List[str]]:
    """
    获取数据存储列表
    :param geoserver： ip_port、username和password配置的字典
    :param workspace: 工作空间名称
    :return: 数据存储列表
    """
    response = requests.get(
        f'http://{geoserver["ip_port"]}/geoserver/rest/workspaces/{workspace}/coveragestores',
        auth=(geoserver["username"], geoserver["password"]),
    )
    if response.status_code == 200:
        response_json = response.json()['coverageStores']
        if response_json:
            coveragestores = response_json['coverageStore']
            coveragestores_name = [str(wp['name']) for wp in coveragestores]
            return coveragestores_name
        else:
            return None
    else:
        print(f'{workspace}：coveragestores查询失败')
        return None



def print_jindu(jindu, total, per: int = 20, dot: int = 5):
    if not per:
        # 若未指定per，则所有进度都输出
        progress_percentage = round(jindu / total * 100, dot)
        print("进度： ", progress_percentage, " %")
    else:
        # 若指定了per，则将进度五等分
        progress_percentage = round(jindu / total * 100, dot)
        steps = [i * (100 / per) for i in range(1, per + 1)]

        for step in steps:
            deta = 1 / total
            if abs(step - progress_percentage) < deta:
                print("进度： ", round(progress_percentage, dot), " %")
            elif progress_percentage == 100:
                print("进度： ", round(progress_percentage, dot), " %")


import zipfile
from pathlib import Path


def unpack_zip(zip_path, new_path=None) -> str:
    """
    解压zip文件
    :param zip_path: zip文件路径
    :param new_path: 解压后的文件存放路径，默认为zip文件所在目录
    :return: 解压后的文件存放路径
    """
    if not new_path:
        new_path = Path(zip_path).parent

    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        for info in zip_ref.infolist():
            # 处理中文文件名乱码问题
            try:
                info.filename = info.filename.encode('cp437').decode('gbk')
            except Exception as e:
                print(f'解压时编码错误：{e}')

            zip_ref.extract(info, new_path)
    return new_path



import os

from ezdxf import recover
from ezdxf.addons import odafc
from ezdxf.addons.drawing import matplotlib

from .Basic import check_folder

# 通过修改win_exec_path的值为自定义安装路径
# https://ezdxf.readthedocs.io/en/stable/addons/drawing.html#module-ezdxf.addons.drawing
odafc.win_exec_path = os.path.join('./', 'ODA File Converter/ODAFileConverter.exe')

def DWG2DXF(input_dwg, outputRoot=None):
    """
    DWG文件转换为DXF文件
    :param input_dwg: 输入的dwg文件路径
    :param outputRoot: 输出路径
    :return:
    """
    if not outputRoot:
        # 获取文件的目录和文件名
        outputRoot, file_name = os.path.split(input_dwg)

    check_folder(outputRoot)

    dxfName = os.path.basename(input_dwg).replace('.dwg', '.dxf')
    output_dxf = os.path.join(outputRoot, dxfName)
    # 转换dwg文件。replace=true允许覆盖已有文件。
    try:
        odafc.convert(input_dwg, output_dxf, version='R2018', replace=True)
        return output_dxf
    except Exception as e:
        print('dwg文件转换失败', e)
        return False


def DXF2PNG(input_dxf, pngName=None, outputRoot=None):
    """
    DXF文件转换为PNG文件
    :param input_dxf: 输入的dxf文件路径
    :param pngName: 输出png文件名称
    :param outputRoot: 输出路径
    :return:
    """
    if not outputRoot:
        # 获取文件的目录和文件名
        outputRoot, file_name = os.path.split(input_dwg)

    check_folder(outputRoot)

    if not pngName:
        pngName = os.path.basename(input_dxf).replace('.', '_') + '.png'
    output_png = os.path.join(outputRoot, pngName)

    doc, auditor = recover.readfile(input_dxf)
    if not auditor.has_errors:
        matplotlib.qsave(doc.modelspace(), output_png, dpi=300)
        return True
    else:
        return False



import geopandas as gpd


def read_geojson_data(features):
    """解析geojson数据，到shapely.geometry对象"""
    geometry_list = []
    # 转换为GeoDataFrame
    # features = json.loads(features)
    gdf = gpd.GeoDataFrame.from_features(features)
    for index, geometry in enumerate(gdf.geometry):
        geometry_list.append({"name": gdf.id[index],
                              'geometry': geometry})
    return geometry_list



import os

import numpy as np
import rasterio
from numpy import inf
from rasterio.mask import mask
from shapely.geometry import Polygon, Point
from typing import Optional

from ..GeoServer import GeoServer


def get_tif_data(tif_path, geometry=None) -> Optional[list[float]]:
    """
    获取 几何对象 对应的栅格数据
    :param tif_path: tif文件路径
    :param geometry: shapely.geometry的几何对象
    :return: 最小值、平均值、最大值的列表
    """
    # 打开.tif文件
    with rasterio.open(tif_path, 'r') as dataset:
        if geometry.type == 'Point':
            geometry: Point
            lon = geometry.x
            lat = geometry.y
            if -180 <= lon <= 180 and -90 <= lat <= 90:
                pass
            else:
                return None

            # 获取数据集的元数据
            transform = dataset.transform  # 变换矩阵，用于坐标转换
            width = dataset.width  # 数据宽度（列数）
            height = dataset.height  # 数据高度（行数）
            # 将经纬度转换为图像的行列索引
            row, col = rasterio.transform.rowcol(transform, lon, lat)

            # 检查行列索引是否在数据范围内
            if 0 <= row < height and 0 <= col < width:
                # 读取该位置的像素值
                pixel_value = float(dataset.read(1, window=((row, row + 1), (col, col + 1)))[0, 0])
                if pixel_value == dataset.nodata:
                    return None
                pixel_value = round(pixel_value, 2)
                return [pixel_value, pixel_value, pixel_value]
            else:
                print("Coordinates are outside the dataset bounds.")
                return None
        elif geometry.type == 'Polygon':
            poly: Polygon = geometry

            try:
                # 裁剪数据
                clipped_data, _ = mask(dataset=dataset, shapes=[poly], nodata=float(inf), filled=True, crop=True)
            except ValueError:
                return None

            # 使用布尔索引选出非零元素
            useful_elements = clipped_data[clipped_data != float(inf)].tolist()
            if useful_elements:
                # 计算非零元素的平均值
                average_elements = float(np.mean(useful_elements))
                max_elements = float(np.max(useful_elements))
                min_elements = float(np.min(useful_elements))

                if dataset.nodata in [min_elements, average_elements, max_elements]:
                    return None

                min_elements, average_elements, max_elements = \
                    round(min_elements, 2), round(average_elements, 2), round(max_elements, 2)
                return [min_elements, average_elements, max_elements]
            else:
                return None


def publish_tif(tif_path=None, fileData=None, workspace=None, layer=None, geoserver=None) -> bool:
    """工作区管理"""
    # 查询现有workspace列表
    workspace_names = GeoServer.get_workspace_list(geoserver)
    if not workspace_names:
        print('workspace列表获取失败')
        return False

    # 如果工作区不存在，则创建workspace
    if workspace not in workspace_names:
        if not GeoServer.create_workspace(geoserver, workspace):
            print(f'{workspace}工作区不存在，创建workspace失败')
            return False
    else:
        print(f'{workspace}工作区已存在')

    """开始处理geoserver发布"""
    tif_name = os.path.basename(tif_path).replace('.tif', '')
    if not layer:
        layer = tif_name

    if GeoServer.publish_tif(geoserver, workspace, layer, tif_path, fileData=fileData):
        # 使用序列化器更新数据
        content_data = {
            'geoserver_state': '1',
            'wmsUrl': f'http://{geoserver["ip_port"]}/geoserver/',
            'workplace': workspace,
            'layer': layer,
            'data_type': '栅格数据'
        }
        return True


def get_min_max_value(tif_path) -> tuple[float, float, float]:
    """
    获取 TIFF 图像的最大值和最小值。

    :param tif_path: TIFF 文件的路径
    :return: 最小值和最大值的元组 (min_value, max_value)
    """
    with rasterio.open(tif_path, 'r') as src:
        # 读取图像数据
        image_data = src.read(1)  # 读取第一个波段的数据

        # 排除 NoData 值
        valid_data = image_data[image_data != src.nodata]

        # 计算最小值和最大值
        min_value = float(valid_data.min())
        max_value = float(valid_data.max())
        noData = float(src.nodata)

    return min_value, max_value, noData



# https://blog.csdn.net/weixin_43931979/article/details/130191956
# https://blog.csdn.net/Tonque/article/details/113615135

import math
from pyproj import Proj
from pyproj import Transformer

pi = 3.1415926535897932384626  # 圆周率π
x_pi = pi * 3000.0 / 180.0  # 圆周率转换量
a = 6378245.0  # 卫星椭球坐标投影到平面地图坐标系的投影因子
ee = 0.00669342162296594323  # 椭球偏心率


def _out_of_china(lng, lat):
    """
    判断是否在国内，不在国内不做偏移
    :param lng:
    :param lat:
    :return:
    """
    return not (73.66 < lng < 135.05 and 3.86 < lat < 53.55)


def _transform_lat(lng, lat):
    ret = -100.0 + 2.0 * lng + 3.0 * lat + 0.2 * lat * lat + 0.1 * lng * lat + 0.2 * math.sqrt(math.fabs(lng))
    ret += (20.0 * math.sin(6.0 * lng * pi) + 20.0 * math.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * math.sin(lat * pi) + 40.0 * math.sin(lat / 3.0 * pi)) * 2.0 / 3.0
    ret += (160.0 * math.sin(lat / 12.0 * pi) + 320 * math.sin(lat * pi / 30.0)) * 2.0 / 3.0
    return ret


def _transform_lng(lng, lat):
    ret = 300.0 + lng + 2.0 * lat + 0.1 * lng * lng + 0.1 * lng * lat + 0.1 * math.sqrt(math.fabs(lng))
    ret += (20.0 * math.sin(6.0 * lng * pi) + 20.0 * math.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * math.sin(lng * pi) + 40.0 * math.sin(lng / 3.0 * pi)) * 2.0 / 3.0
    ret += (150.0 * math.sin(lng / 12.0 * pi) + 300.0 * math.sin(lng / 30.0 * pi)) * 2.0 / 3.0
    return ret


def wgs84_to_gcj02(lng, lat):
    """
    WGS84转火星坐标系(GCJ - 02)
    :param lng: WGS84坐标系的经度
    :param lat: WGS84坐标系的纬度
    :return:
    """
    # 判断是否在国内
    if _out_of_china(lng, lat):
        return lng, lat
    d_lat = _transform_lat(lng - 105.0, lat - 35.0)
    d_lng = _transform_lng(lng - 105.0, lat - 35.0)
    rad_lat = lat / 180.0 * pi
    magic = math.sin(rad_lat)
    magic = 1 - ee * magic * magic
    sqrt_magic = math.sqrt(magic)
    d_lat = (d_lat * 180.0) / ((a * (1 - ee)) / (magic * sqrt_magic) * pi)
    d_lng = (d_lng * 180.0) / (a / sqrt_magic * math.cos(rad_lat) * pi)
    mg_lat = lat + d_lat
    mg_lng = lng + d_lng
    return mg_lng, mg_lat


def gcj02_to_wgs84(lng, lat):
    """
    GCJ02(火星坐标系)转WGS84
    :param lng: 火星坐标系的经度
    :param lat: 火星坐标系纬度
    :return:
    """
    # 判断是否在国内
    if _out_of_china(lng, lat):
        return lng, lat
    d_lat = _transform_lat(lng - 105.0, lat - 35.0)
    d_lng = _transform_lng(lng - 105.0, lat - 35.0)
    rad_lat = lat / 180.0 * pi
    magic = math.sin(rad_lat)
    magic = 1 - ee * magic * magic
    sqrt_magic = math.sqrt(magic)
    d_lat = (d_lat * 180.0) / ((a * (1 - ee)) / (magic * sqrt_magic) * pi)
    d_lng = (d_lng * 180.0) / (a / sqrt_magic * math.cos(rad_lat) * pi)
    mg_lat = lat + d_lat
    mg_lng = lng + d_lng
    return [lng * 2 - mg_lng, lat * 2 - mg_lat]


def from_geo_to_proj(lon, lat, crs_id=None):
    """
    只能从geo向proj转换，其中proj是CGCS2000系列
    无法进行带有带号的转换
    """
    if crs_id:
        pass
    else:
        crs_base = 4534
        if 73.5 < lon < 136.5:
            result = divmod(lon - 75, 3)

            shang = result[0]
            yu = result[1]

            if yu < 1.5:
                num = shang
            else:
                num = shang + 1
            crs_id = crs_base + num

        elif lon < 73.5:
            crs_id = 4534

        else:
            crs_id = 4554

    # EPSG = 'epsg:' + str(int(crs_id))
    # proj = Proj(EPSG)

    proj = Proj(int(crs_id))
    x, y = proj(lon, lat)

    return x, y, crs_id

def from_proj_to_geo(x, y, crs_id=None):
    # https://pyproj4.github.io/pyproj/stable/examples.html#transformations-from-crs-to-crs

    # 目标地理坐标系，GCS_China_Geodetic_Coordinate_System_2000，WKID: 4490 权限: EPSG
    transformer = Transformer.from_crs(int(crs_id), 4490, always_xy=True)

    # 转换到经纬度
    lon, lat = transformer.transform(x, y)
    return lon, lat




import os
import subprocess

subprocess.run(['pip', 'freeze'])

# 将包列表保存到文件
with open('packages.txt', 'w') as f:
    subprocess.run(['pip', 'freeze'], stdout=f)

# 读取文件中的包列表
with open('packages.txt') as f:
    packages = f.read().splitlines()

# 卸载每个包
for package in packages:
    package = package.split('==')[0]
    if package in ['pip', 'setuptools']:
        continue
    print(f"Uninstalling {package}...")
    subprocess.run(['pip', 'uninstall', '-y', package])

os.remove('packages.txt')



from setuptools import setup, find_packages
# python setup.py sdist bdist_wheel
# twine upload dist/* --verbose

setup(
    name='hcapy',
    version='1.7',
    packages=find_packages(),
    description='HCA private package',
    author='huochunan',
    author_email='huochunan@163.com',
    license='MIT',
    install_requires=[
        'ezdxf',
        'geopandas',
        'pyshp',
        'rasterio',
        'numpy',
        'pyproj'
    ],
    classifiers=[
        # 分类信息
    ]
)



import os


def check_folder(folder_path):
    """检查文件夹是否存在，不存在则创建"""
    # 检查文件夹是否存在
    if not os.path.exists(folder_path):
        # 如果不存在，创建文件夹
        os.makedirs(folder_path)
        # 创建成功后，你可以在这里添加一些其他操作
        print(f"文件夹'{folder_path}'已创建。")
    else:
        # 文件夹已经存在，你可以选择在这里做其他处理
        print(f"文件夹'{folder_path}'已存在。")


def get_subdirectories_and_files(directory) -> tuple[list, list]:
    """
    获取子目录和子文件
    :param directory: 目录路径
    :return: 子目录列表，子文件列表
    """
    subdirectories = []
    files = []
    for root, dirs, filenames in os.walk(directory):
        for DIR in dirs:
            subdirectories.append(os.path.join(root, DIR))
        for file in filenames:
            files.append(os.path.join(root, file))
    # 将目录的结构文件写道目录所在路径，避免对目录结构的影响
    # base_path = upper_dictionary(directory)
    base_path = directory

    # 写入子目录的绝对地址到txt文件
    with open(os.path.join(base_path, 'subdirectories.txt'), 'w', encoding='utf-8') as file:
        file.write('\n'.join(subdirectories))

    # 写入子文件的绝对地址到txt文件
    with open(os.path.join(base_path, 'files.txt'), 'w', encoding='utf-8') as file:
        file.write('\n'.join(files))

    return subdirectories, files


def upper_dictionary(dictionary, num):
    """
    获取文件绝对地址的父地址
    :param dictionary: 文件绝对地址
    :param num: 父级目录的层数
    :return: 父级目录的绝对地址
    """
    parent_dir = dictionary
    if num <= 0:
        pass
    else:
        for i in range(num):
            parent_dir = os.path.abspath(os.path.join(dictionary, ".."))
            dictionary = parent_dir
    return parent_dir
